@@ -1,4 +1,4 @@
-from typing import Any, List, Union, Dict
+from typing import Any, List, Union, Dict, Tuple, Optional
 from typing_extensions import TypedDict, Literal
 
 DataSource = Literal["issues", "boards"]
@@ -32,13 +32,14 @@ QueryDefinition = TypedDict(
         "from": DataSource,
         "where": JqlList,
         "order_by": JqlList,
+        "filter": ExpressionList,
         "having": ExpressionList,
         "group_by": ExpressionList,
         "sort_by": ExpressionList,
         "expand": List[str],
         "limit": int,
         "cap": int,
-        "cache": int,
+        "cache": Union[int, Tuple[Optional[int], Optional[int]]],
     },
     total=False,
 )
