@@ -8,7 +8,9 @@ from types import ModuleType
 from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union, Tuple
 
 from appdirs import user_config_dir
+from jira.resources import Resource
 import simpleeval
+from simpleeval import EvalWithCompoundTypes
 from yaml import safe_dump, safe_load
 
 from .constants import APP_NAME
@@ -16,7 +18,6 @@ from .exceptions import QueryError
 from .types import (
     ConfigDict,
     ExpressionList,
-    QueryDefinition,
     SelectFieldDefinition,
     Expression,
 )
@@ -39,6 +40,14 @@ def get_config_dir() -> str:
     return root_path
 
 
+def get_cache_path() -> str:
+    config_path = get_config_dir()
+    cache_path = os.path.join(config_path, "cache")
+    os.makedirs(cache_path, exist_ok=True)
+
+    return cache_path
+
+
 def get_default_config_path() -> str:
     root_path = get_config_dir()
     return os.path.join(root_path, "config.yaml",)
@@ -81,36 +90,6 @@ def get_functions_for_module(
     return functions
 
 
-def clean_query_definition(query: QueryDefinition) -> QueryDefinition:
-    cleaned_query: QueryDefinition = {
-        "select": [],
-        "from": query["from"],
-    }
-
-    for select in query["select"]:
-        if isinstance(select, dict):
-            cleaned_query["select"].append(select)
-        else:
-            cleaned_query["select"].append(str(select))
-
-    for section in (
-        "where",
-        "order_by",
-        "having",
-        "group_by",
-        "sort_by",
-        "expand",
-    ):
-        if section in query:
-            # I'm not sure I understand why the typing checks here are failing
-            cleaned_query[section] = [str(line) for line in query[section]]  # type: ignore
-
-    if "limit" in query:
-        cleaned_query["limit"] = query["limit"]
-
-    return cleaned_query
-
-
 def parse_select_definition(
     expression: Union[str, SelectFieldDefinition]
 ) -> SelectFieldDefinition:
@@ -151,26 +130,22 @@ def calculate_result_hash(
     return int(hashlib.sha1(":".join(params).encode("UTF-8")).hexdigest(), 16)
 
 
-def field_name_is_public(key: str) -> bool:
-    if key.lower() == key and key != "fields" and not key.startswith("_"):
-        return True
-
-    return False
-
-
 def get_row_dict(row: Any) -> Dict[str, Any]:
     names: Dict[str, Any] = {}
 
     if hasattr(row, "fields"):
         for field_name in dir(row.fields):
-            if field_name_is_public(field_name):
+            if not field_name.startswith("_"):
                 names[field_name] = getattr(row.fields, field_name)
 
     # Gather any top-level keys, too, to make sure we fetch any expansions
     for key in dir(row):
         value = getattr(row, key)
-        if field_name_is_public(key) and isinstance(
-            value, (str, bool, float, int, list, dict)
+        if (
+            key != "fields"
+            and not key.startswith("_")
+            and not key.upper() == key
+            and not callable(value)
         ):
             names[key] = getattr(row, key)
 
@@ -192,7 +167,37 @@ def evaluate_expression(
     names: Dict[str, Any],
     functions: Optional[Dict[str, Callable]] = None,
 ) -> Any:
-    return simpleeval.simple_eval(expression, names=names, functions=functions)
+    return EvalWithCompoundTypes(names=names, functions=functions).eval(expression)
+
+
+def normalize_value(value: Any) -> Any:
+    if isinstance(value, (str, bool, float, int)):
+        return value
+    elif isinstance(value, list):
+        return [normalize_value(v) for v in value]
+    elif isinstance(value, Dict):
+        return {k: normalize_value(v) for k, v in value.items()}
+    elif isinstance(value, Resource):
+        display = str(value)
+        if display.startswith("<JIRA"):
+            # If the display value starts with "<JIRA", we know
+            # the library couldn't find a pretty way of printing
+            # this field; let's just return the raw dictionary
+            # in that case
+            return value.raw
+        return display
+    elif "PropertyHolder" in str(value):
+        # Yes; this LOOKS ridiculous, but in older versions
+        # of the Jira library, the PropertyHolder class is
+        # created a runtime :facepalm:.  I think this is
+        # the only way to find them :shrug:
+        return {
+            name: normalize_value(getattr(value, name))
+            for name in dir(value)
+            if not name.startswith("_") and not name.upper() == name
+        }
+
+    return value
 
 
 def get_field_data(
@@ -205,11 +210,14 @@ def get_field_data(
         functions = {}
 
     try:
-        return simpleeval.simple_eval(
-            expression, names={"_": row, **row.as_dict()}, functions=functions
+        return normalize_value(
+            evaluate_expression(
+                expression, names={"_": row, **row.as_dict()}, functions=functions,
+            )
         )
     except (
         simpleeval.AttributeDoesNotExist,
+        simpleeval.NameNotDefined,
         KeyError,
         IndexError,
         TypeError,
@@ -218,5 +226,5 @@ def get_field_data(
         if not error_returns_null:
             raise
         return None
-    except simpleeval.NameNotDefined as e:
+    except Exception as e:
         raise QueryError(str(e))
